from pylibvirt.modules.devices import Device


class RngDevice(Device):
    XML_NAME = "rng"

    def __init__(self, model: str = 'virtio', backend_model: str = "random",
                 host_device: str = '/dev/urandom'):
        super().__init__(name=self.XML_NAME)
        self.__model = model
        self.__backend_model = backend_model
        self.__host_device = host_device
        self.generate_data()

    @property
    def model(self) -> str:
        return self.__model

    @model.setter
    def model(self, model: str):
        self.__model = model

    @property
    def backend_model(self) -> str:
        return self.__backend_model

    @backend_model.setter
    def backend_model(self, backend_model: str):
        self.__backend_model = backend_model

    @property
    def host_device(self) -> str:
        return self.__host_device

    @host_device.setter
    def host_device(self, host_device: str):
        self.__host_device = host_device

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "model": self.model,
                },
                "children": {
                    "backend": {
                        "text": self.host_device,
                        "attr": {
                            "model": self.backend_model
                        }
                    }
                }
            }
        })



from pylibvirt.modules.devices import Device


class VcpuDevice(Device):
    XML_NAME = "vcpu"

    def __init__(self, vcpu: int, vcpu_args=None):
        super().__init__(name=self.XML_NAME)
        if vcpu_args is None:
            vcpu_args = {'placement': 'static'}
        self.root = []
        self.__vcpu_args = vcpu_args
        self.__vcpu = vcpu
        self.generate_data()

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": self.__vcpu_args,
                "text": str(self.__vcpu)
            }
        })


class CpuDevice(Device):
    XML_NAME = "cpu"

    def __init__(self, vcpu_args: dict or None = None, cpu_model: str = 'host',
                 model_args=None, cpu_args=None, vcpu: int = 1,
                 topology_args: dict or None =
                 None):
        super().__init__(name=self.XML_NAME)
        if model_args is None:
            model_args = {'fallback': 'allow'}
        if cpu_args is None:
            cpu_args = {'mode': 'host-model', 'check': 'partial'}
        self.root = []
        self.__cpu_model = cpu_model
        self.__cpu_args = cpu_args
        self.__model_args = model_args
        self.__topology = topology_args
        if not self.__topology:
            self.__vcpu = VcpuDevice(vcpu=vcpu, vcpu_args=vcpu_args)
        else:
            self.__vcpu = VcpuDevice(vcpu=self.count_vcpu(), vcpu_args=vcpu_args)
        self.generate_data()

    @property
    def cpu_model(self) -> str:
        return self.__cpu_model

    @cpu_model.setter
    def cpu_model(self, cpu_model: str):
        self.__cpu_model = cpu_model

    @property
    def cpu_args(self) -> dict:
        return self.__cpu_args

    @cpu_args.setter
    def cpu_args(self, cpu_args: str):
        self.__cpu_args = cpu_args

    @property
    def topology(self):
        if not self.__topology:
            return False
        elif 'sockets' not in self.__topology or 'cores' not in self.__topology or \
                'threads' not in self.__topology:
            return False
        else:
            return self.__topology

    def count_vcpu(self):
        return self.topology['sockets'] * self.topology['cores'] * self.topology[
            'threads']

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": self.cpu_args,
                "children": {
                    "model": {
                        "text": self.cpu_model
                    },
                    "attr": self.__model_args
                }
            }
        })
        self.data.update(self.__vcpu.data)



import xml.etree.cElementTree as ET  # nosec
import defusedxml


class XmlGenerator:
    def __init__(self):
        defusedxml.defuse_stdlib()

    def __create_xml_elt(self, data, parent):
        for key, value in data.items():
            attr = value.get('attr', {})
            text = value.get('text', None)
            children = value.get('children', None)

            entry = ET.SubElement(parent, key, attrib=attr)
            if text:
                entry.text = text

            if children:
                self.add_node(data=children, parent=entry)

    def add_node(self, data, parent):
        if isinstance(data, list):
            for d in data:
                self.__create_xml_elt(data=d, parent=parent)
        else:
            self.__create_xml_elt(data=data, parent=parent)

    def generate_xml_element(self, data: dict):
        for key, value in data.items():
            attr = value.get('attr', {})
            text = value.get('text', None)
            children = value.get('children', None)

            document = ET.Element(key, attrib=attr)
            if text:
                document.text = text
            if children:
                self.add_node(data=children, parent=document)

        return ET.tostring(document).decode()



from pylibvirt.modules.devices import Device


class Storage(Device):
    XML_NAME = "pool"

    def __init__(self, pool_type: str, pool_name: str):
        super().__init__(name=self.XML_NAME)
        self.__pool_type = pool_type
        self.__pool_name = pool_name

    @property
    def pool_name(self) -> str:
        return self.__pool_name

    @pool_name.setter
    def pool_name(self, pool_name: str):
        self.__pool_name = pool_name

    @property
    def pool_type(self) -> str:
        return self.__pool_type

    @pool_type.setter
    def pool_type(self, pool_type: str):
        self.__pool_type = pool_type

    def generate_data(self):
        self.data[self.xml_name].update({"children": {
            "name": {
                "text": self.pool_name
            }
        }})
        self.data[self.xml_name].update({"attr": {
            "type": self.pool_type
        }})


class DirStorage(Storage):
    def __init__(self, pool_name: str, path: str):
        super().__init__(pool_type="dir", pool_name=pool_name)
        self.__path = path
        self.generate_data()

    @property
    def path(self) -> str:
        return self.__path

    @path.setter
    def path(self, path: str):
        self.__path = path

    def add_path(self):
        return {
            "target": {
                "children": {
                    "path": {
                        "text": self.path
                    }
                }
            }
        }

    def generate_data(self):
        super().generate_data()
        self.update_data(self.add_path())



from pylibvirt.modules.devices import Device


class VideoDevice(Device):
    XML_NAME = "video"

    def __init__(self, model_type: str = "qxl", ram: int = 65536, vram: int = 65536,
                 vgamem: int = 16384):
        super().__init__(name=self.XML_NAME)
        self.__model = model_type
        self.__ram = ram
        self.__vram = vram
        self.__vgamem = vgamem
        self.generate_data()

    @property
    def model(self) -> str:
        return self.__model

    @model.setter
    def model(self, model: str):
        self.__model = model

    @property
    def vgamem(self) -> int:
        return self.__vgamem

    @vgamem.setter
    def vgamem(self, vgamem: int):
        self.__vgamem = vgamem

    @property
    def ram(self) -> int:
        return self.__ram

    @ram.setter
    def ram(self, ram: int):
        self.__ram = ram

    @property
    def vram(self) -> int:
        return self.__vram

    @vram.setter
    def vram(self, vram: int):
        self.__vram = vram

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "children": {
                    "model": {
                        "attr": {
                            "type": self.model,
                            "ram": str(self.ram),
                            "vram": str(self.vram),
                            "vgamem": str(self.vgamem),
                        }
                    }
                }
            }
        })



from defusedxml import minidom

import libvirt

from pylibvirt.modules.devices import Device

DEFAULT_BUS = 'virtio'


class DiskDevice(Device):
    XML_NAME = "disk"

    def __init__(self, volume: libvirt.virStorageVol, target: str, driver: str = 'qemu',
                 bus: str = DEFAULT_BUS,
                 disk_type: str = "file",
                 device: str = "disk",
                 cache: str = "default", discard: str = "default",
                 detect_zeroes: str = "default", shareable: bool = False,
                 readonly: bool = False):
        super().__init__(name=self.XML_NAME)
        self.__volume = volume
        self.__target = target
        self.__driver = driver
        self.__bus = bus
        self.__disk_type = disk_type
        self.__device = device
        self.__cache = cache
        self.__discard = discard
        self.__detect_zeroes = detect_zeroes
        self.__shareable = shareable
        self.__readonly = readonly
        self.__format = None
        self.__path = None
        self.__disk_info()
        self.generate_data()

    @property
    def path(self) -> str:
        return self.__path

    @path.setter
    def path(self, path: str):
        self.__path = path

    @property
    def format(self) -> str:
        return self.__format

    @format.setter
    def format(self, format: str):
        self.__format = format

    @property
    def driver(self) -> str:
        return self.__driver

    @driver.setter
    def driver(self, driver: str):
        self.__driver = driver

    @property
    def volume(self) -> libvirt.virStorageVol:
        return self.__volume

    @property
    def readonly(self) -> bool:
        return self.__readonly

    @readonly.setter
    def readonly(self, readonly: bool):
        self.__readonly = readonly

    @property
    def shareable(self) -> bool:
        return self.__shareable

    @shareable.setter
    def shareable(self, shareable: bool):
        self.__shareable = shareable

    @property
    def detect_zeroes(self) -> str:
        return self.__detect_zeroes

    @detect_zeroes.setter
    def detect_zeroes(self, detect_zeroes: str):
        self.__detect_zeroes = detect_zeroes

    @property
    def discard(self) -> str:
        return self.__discard

    @discard.setter
    def discard(self, discard: str):
        self.__discard = discard

    @property
    def cache(self) -> str:
        return self.__cache

    @cache.setter
    def cache(self, cache: str):
        self.__cache = cache

    @property
    def device(self) -> str:
        return self.__device

    @device.setter
    def device(self, device: str):
        self.__device = device

    @property
    def disk_type(self) -> str:
        return self.__disk_type

    @disk_type.setter
    def disk_type(self, disk_type: str):
        self.__disk_type = disk_type

    @property
    def bus(self) -> str:
        return self.__bus

    @bus.setter
    def bus(self, bus: str):
        self.__bus = bus

    @property
    def target(self) -> str:
        return self.__target

    @target.setter
    def target(self, target: str):
        self.__target = target

    def __disk_info(self):
        xml = self.volume.XMLDesc(0)
        xml = minidom.parseString(xml)
        xml_path = xml.getElementsByTagName('path')[0]
        xml_format = xml.getElementsByTagName('format')[0]
        self.path = xml_path.firstChild.data
        _format = xml_format.getAttribute('type')
        if _format == "iso":
            self.device = "cdrom"
            self.format = "raw"
            self.bus = "sata"
        else:
            self.format = _format

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "type": self.disk_type,
                    "device": self.device
                },
                "children": {
                    "driver": {
                        "attr": {
                            "name": self.driver,
                            "type": self.format
                        }
                    },
                    "source": {
                        "attr": {
                            "file": self.path,
                        }
                    },
                    "target": {
                        "attr": {
                            "dev": self.target,
                            "bus": self.bus
                        },
                    }
                }
            }
        })
        if self.readonly:
            self.update_data({
                "readonly": {}
            })
        if self.shareable:
            self.update_data({
                "shareable": {}
            })



import string
import uuid
from defusedxml import minidom

from pylibvirt.modules.devices.device import Device


def generate_dev(patter: str):
    alpha = string.ascii_lowercase
    dev_list = []
    for letter in alpha:
        dev_list.append(patter + letter)
    return dev_list


def next_dev(dev_used: list, dev: str = 'vd'):
    dev_used = set(dev_used)
    dev_list = set(generate_dev(dev))
    return sorted(list(dev_list - dev_used))


def next_dev_from_dict(dom_xml: str, dev: str = 'vd'):
    xml = minidom.parseString(dom_xml)
    disks = xml.getElementsByTagName('disk')
    dev_used = []
    for disk in disks:
        for target in disk.getElementsByTagName('target'):
            dev_used.append(target.getAttribute('dev'))
    dev_used = set(dev_used)
    dev_list = set(generate_dev(dev))
    return sorted(list(dev_list - dev_used))


def get_dev(bus: str):
    if bus == 'scsi':
        return 'sd'
    elif bus == 'virtio':
        return 'vd'
    elif bus == 'usb':
        return 'sd'
    elif bus == 'sata':
        return 'sd'
    elif bus == 'fdc':
        return 'fd'


def _add_sub_feature(feature):
    child = {}
    for key, data in feature.items():
        if isinstance(feature, str):
            child.update({key: {}})
        elif isinstance(feature, dict):
            child.update({key: {
                "attr": data
            }})
    return child


class Feature(Device):
    XML_NAME = "features"

    def __init__(self, feature=None):
        super().__init__(name=self.XML_NAME)
        self.root = []
        if feature is None:
            feature = ['acpi', 'acpi']
        self._feature = feature
        self.generate_data()

    def _convert_feature(self):
        for feature in self._feature:
            if isinstance(feature, str):
                self.update_data(data={feature: {}})
            elif isinstance(feature, dict):
                for key, data in feature.items():
                    self.update_data(data={key: {"children": _add_sub_feature(data)}})

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "children": {
                }
            }
        })
        self._convert_feature()


class Domain(Device):
    XML_NAME = "domain"

    def __init__(self, name: str, domain_type: str = "kvm", devices=None,
                 boot_order=None, os: dict = None, feature: list = None):
        super().__init__(name=self.XML_NAME)
        if os is None:
            os = {'arch': 'x86_64', 'machine': 'q35', 'os_type': 'hvm'}
        if devices is None:
            devices = []
        if boot_order is None:
            boot_order = ['network', 'cdrom', 'hd']
        self.__domain_type = domain_type
        self.__feature = feature
        self.__uuid = str(uuid.uuid4())
        self.__name = name
        self.__devices = devices
        self.__boot_order = boot_order
        self.__os = os
        self.__memory = self.set_memory()
        if feature:
            self.__devices.append(self.get_feature())
        self.generate_data()

    @property
    def boot_order(self) -> list:
        return self.__boot_order

    @boot_order.setter
    def boot_order(self, boot_order: list):
        self.__boot_order = boot_order

    @property
    def devices(self) -> list:
        return self.__devices

    @devices.setter
    def devices(self, devices: list):
        self.__devices = devices

    @property
    def domain_type(self) -> str:
        return self.__domain_type

    @domain_type.setter
    def domain_type(self, domain_type: str):
        self.__domain_type = domain_type

    @property
    def uuid(self) -> str:
        return self.__uuid

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name

    @property
    def memory(self) -> dict:
        return self.__memory

    @memory.setter
    def memory(self, memory: {}):
        self.__memory = memory
        self.generate_data()

    @staticmethod
    def set_memory(memory: int = 1, max_memory: int = 2, mem_unit: str = "G"):
        return {"memory": {
            "text": str(max_memory),
            "attr": {
                "unit": mem_unit
            }
        },
            "currentMemory": {
                "text": str(memory),
                "attr": {
                    "unit": mem_unit
                }
            }
        }

    @property
    def os(self) -> dict:
        return self.__os

    def get_os(self) -> dict:
        data = {
            "os": {
                "children": [{
                    "type": {
                        "attr": {
                            "arch": self.os.get('arch'),
                            "machine": self.os.get('machine'),
                        },
                        "text": self.os.get('os_type')
                    }
                }]
            }
        }
        boot_section = data["os"]["children"]
        for boot in self.boot_order:
            boot_section.append({"boot": {
                "attr": {
                    "dev": boot
                }
            }})

        return data

    def get_feature(self):
        feature = Feature(feature=self.__feature)
        return feature

    def add_device(self, device: Device):
        self.devices.append(device)

    def add_devices_to_data(self, device: Device):
        device.generate_data()
        devices = self.data[self.XML_NAME]
        if len(device.root) > 0:
            for root in device.root:
                devices = devices["children"][root]
            devices = devices["children"]
            devices.append(device.data)
        else:
            devices = devices["children"]
            devices.update(device.data)

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "type": self.domain_type
                },
                "children": {
                    "name": {
                        "text": self.name
                    },
                    "uuid": {
                        "text": self.uuid
                    },
                    "devices": {
                        "children": [

                        ]
                    }
                }
            }
        })
        self.update_data(self.get_os())
        self.update_data(self.memory)

        for device in self.devices:
            self.add_devices_to_data(device)



from pylibvirt.modules.devices import Device


class FileSystemDevice(Device):
    XML_NAME = "filesystem"

    def __init__(self, source: str, target: str, mount_type: str = "mount",
                 access_mode: str = "mapped",
                 read_only: bool = False):
        super().__init__(name=self.XML_NAME)
        self.__source = source
        self.__target = target
        self.__mount_type = mount_type
        self.__access_mode = access_mode
        self.__read_only = read_only
        self.generate_data()

    @property
    def read_only(self) -> bool:
        return self.__read_only

    @read_only.setter
    def read_only(self, read_only: bool):
        self.__read_only = read_only

    @property
    def access_mode(self) -> str:
        return self.__access_mode

    @access_mode.setter
    def access_mode(self, access_mode: str):
        self.__access_mode = access_mode

    @property
    def mount_type(self) -> str:
        return self.__mount_type

    @mount_type.setter
    def mount_type(self, mount_type: str):
        self.__mount_type = mount_type

    @property
    def target(self) -> str:
        return self.__target

    @target.setter
    def target(self, target: str):
        self.__target = target

    @property
    def source(self) -> str:
        return self.__source

    @source.setter
    def source(self, source: str):
        self.__source = source

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "type": self.mount_type,
                    "accessmode": self.access_mode
                },
                "children": {
                    "source": {
                        "attr": {
                            "dir": self.source
                        }
                    },
                    "target": {
                        "attr": {
                            "dir": self.target
                        }
                    }
                }
            }
        })
        # TODO: if self.read_only add readonly



from .device import *
from .rng import *
from .usb import *
from .xml import *
from .sound import *
from .controller import *
from .file_system import *
from .graphic import *
from .interface import *
from .video import *
from .strorage import *
from .cpu import *



import libvirt

from pylibvirt.modules.devices import Device


def create_disk(storage: libvirt.virStoragePool, capacity: int, name: str,
                disk_format: str = 'raw', size_unit: str = 'G', allocation: int = 0):
    new_volume = Volume(name=name, capacity=capacity, disk_format=disk_format,
                        size_unit=size_unit,
                        allocation=allocation)
    return storage.createXML(new_volume.generate_xml(), 0)


def clone_disk(storage: libvirt.virStoragePool, volume: libvirt.virStorageVol,
               capacity: int, name: str,
               disk_format: str = 'raw', size_unit: str = 'G', allocation: int = 0):
    new_volume = Volume(name=name, capacity=capacity, disk_format=disk_format,
                        size_unit=size_unit,
                        allocation=allocation)
    return storage.createXMLFrom(new_volume.generate_xml(), volume, 0)


class Volume(Device):
    XML_NAME = "volume"

    def __init__(self, name: str, capacity: int, allocation: int = 0,
                 size_unit: str = 'G', disk_format: str = "raw"):
        super().__init__(name=self.XML_NAME)
        self.__name = name
        self.__capacity = capacity
        self.__allocation = allocation
        self.__size_unit = size_unit
        self.__disk_format = disk_format
        self.generate_data()

    @property
    def disk_format(self) -> str:
        return self.__disk_format

    @disk_format.setter
    def disk_format(self, disk_format: str):
        self.__disk_format = disk_format

    @property
    def size_unit(self) -> str:
        return self.__size_unit

    @size_unit.setter
    def size_unit(self, size_unit: str):
        self.__size_unit = size_unit

    @property
    def allocation(self) -> int:
        return self.__allocation

    @allocation.setter
    def allocation(self, allocation: int):
        self.__allocation = allocation

    @property
    def capacity(self) -> int:
        return self.__capacity

    @capacity.setter
    def capacity(self, capacity: int):
        self.__capacity = capacity

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name

    def generate_data(self):
        self.data[self.xml_name].update({"children": {
            "name": {
                "text": self.name
            },
            "allocation": {
                "text": str(self.allocation),
                "attr": {
                    "unit": self.size_unit
                }
            },
            "capacity": {
                "text": str(self.capacity),
                "attr": {
                    "unit": self.size_unit
                }
            },
            "target": {
                "children": {
                    "format": {
                        "attr": {
                            "type": self.disk_format
                        }
                    }
                }
            }

        }})



import secrets

from pylibvirt.modules.devices import Device


class NetworkInterfaceDevice(Device):
    XML_NAME = "interface"

    def __init__(self, net_type: str = 'network', net_interface: str = 'default',
                 model: str = 'virtio',
                 mac: str = None):
        super().__init__(name=self.XML_NAME)
        self.__net_type = net_type
        self.__net_interface = net_interface
        self.__model = model
        if not mac:
            self.mac = self.random_mac()
        else:
            self.__mac = mac
        self.generate_data()

    @staticmethod
    def random_mac():
        return "02:00:00:%02x:%02x:%02x" % (secrets.randbelow(255),
                                            secrets.randbelow(255),
                                            secrets.randbelow(255))

    @property
    def net_type(self) -> str:
        return self.__net_type

    @net_type.setter
    def net_type(self, model: str):
        self.__net_type = model

    @property
    def mac(self) -> str:
        return self.__mac

    @mac.setter
    def mac(self, mac: str):
        self.__mac = mac

    @property
    def model(self) -> str:
        return self.__model

    @model.setter
    def model(self, model: str):
        self.__model = model

    @property
    def net_interface(self) -> str:
        return self.__net_interface

    @net_interface.setter
    def net_interface(self, net_interface: str):
        self.__net_interface = net_interface

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "type": self.net_type
                },
                "children": {
                    "source": {
                        "attr": {
                            self.net_type: self.net_interface
                        }
                    },
                    "mac": {
                        "attr": {
                            "address": str(self.mac)
                        }
                    },
                    "model": {
                        "attr": {
                            "type": self.model
                        }
                    }
                }
            }
        })



import ipaddress

from pylibvirt.modules.devices import Device


class Network(Device):
    XML_NAME = "network"

    def __init__(self, name: str, mode: str or bool = "nat", domain: str = "network",
                 ip4_cidr: str = None,
                 dhcp_start: str = None,
                 dhcp_stop: str = None):
        """
        Create virtual network
        :param name: name of the network, must be unique
        :param mode: Type of network, values can be nat, route, open,
        hostdev and False for isolated network
        :param domain: DNS domain name to use, default is network
        :param ip4_cidr: Optional parameter to set the ip configuration:
        eg: 192.168.1.0/24
        :param dhcp_start: Optional parameter to activate dhcp set start address
        and stop address. eg: 192.168.1.2
        :param dhcp_stop: Optional parameter to activate dhcp set start address
         and stop address. eg: 192.168.1.20
        """
        super().__init__(name=self.XML_NAME)
        self.__name = name
        self.__mode = mode
        self.__domain = domain
        self.__ip4_cidr = ip4_cidr
        self.__dhcp_start = dhcp_start
        self.__dhcp_stop = dhcp_stop
        self.generate_data()

    @property
    def dhcp_stop(self) -> str:
        return self.__dhcp_stop

    @dhcp_stop.setter
    def dhcp_stop(self, dhcp_stop: str):
        self.__dhcp_stop = dhcp_stop

    @property
    def dhcp_start(self) -> str:
        return self.__dhcp_start

    @dhcp_start.setter
    def dhcp_start(self, dhcp_start: str):
        self.__dhcp_start = dhcp_start

    @property
    def ip4_cidr(self) -> str:
        return self.__ip4_cidr

    @ip4_cidr.setter
    def ip4_cidr(self, ip4_cidr: str):
        self.__ip4_cidr = ip4_cidr

    @property
    def mode(self) -> str:
        return self.__mode

    @mode.setter
    def mode(self, mode: str):
        self.__mode = mode

    @property
    def domain(self) -> str:
        return self.__domain

    @domain.setter
    def domain(self, domain: str):
        self.__domain = domain

    @property
    def name(self) -> str:
        return self.__name

    @name.setter
    def name(self, name: str):
        self.__name = name

    def add_name(self):
        data = {"name": {
            "text": self.name
        }}
        return data

    def add_domain(self):
        data = {"domain": {
            "attr": {
                "name": self.domain
            }
        }}
        return data

    def add_forward(self):
        data = {"forward": {
            "attr": {
                "mode": self.mode
            }
        }}
        return data

    def add_dhcp(self):
        data = {"dhcp": {
            "children": {
                "range": {
                    "attr": {
                        "start": self.dhcp_start,
                        "end": self.dhcp_stop
                    }
                }
            }
        }}
        return data

    def add_ip(self):
        ip_net = ipaddress.ip_network(self.ip4_cidr)
        netmask = str(ip_net.netmask)
        ip = str(ip_net[1]).split('/')[0]
        data = {"ip": {
            "attr": {
                "address": ip,
                "netmask": netmask
            },
            "children": {

            }
        }}
        return data

    def generate_data(self):
        self.data[self.xml_name].update({"children": {}})
        self.update_data(self.add_name())
        self.update_data(self.add_domain())

        if self.mode:
            self.update_data(self.add_forward())

        if self.ip4_cidr:
            self.update_data(self.add_ip())
            if self.dhcp_start and self.dhcp_stop:
                ip = self.data[self.XML_NAME]["children"]["ip"]["children"]
                ip.update(self.add_dhcp())



from pylibvirt.modules.devices import Device


class GraphicDevice(Device):
    XML_NAME = "graphics"

    def __init__(self, graphic_type: str = "spice", port: int = -1, tls_port: int = -1,
                 password: str = None,
                 image_compression: str = 'off', open_gl: bool = False,
                 render_node: str = "", listen: str = "address"):
        super().__init__(name=self.XML_NAME)
        self.__graphic_type = graphic_type
        self.__port = port
        self.__tls_port = tls_port
        self.__password = password
        self.__image_compression = image_compression
        self.__open_gl = open_gl
        self.__render_node = render_node
        self.__listen = listen
        self.generate_data()

    @property
    def listen(self) -> str:
        return self.__listen

    @listen.setter
    def listen(self, listen: str):
        self.__listen = listen

    @property
    def render_node(self) -> str:
        return self.__render_node

    @render_node.setter
    def render_node(self, render_node: str):
        self.__render_node = render_node

    @property
    def open_gl(self) -> bool:
        return self.__open_gl

    @open_gl.setter
    def open_gl(self, open_gl: bool):
        self.__open_gl = open_gl

    @property
    def image_compression(self) -> str:
        return self.__image_compression

    @image_compression.setter
    def image_compression(self, image_compression: str):
        self.__image_compression = image_compression

    @property
    def password(self) -> str:
        return self.__password

    @password.setter
    def password(self, password: str):
        self.__password = password

    @property
    def tls_port(self) -> int:
        return self.__tls_port

    @tls_port.setter
    def tls_port(self, tls_port: int):
        self.__tls_port = tls_port

    @property
    def port(self) -> int:
        return self.__port

    @port.setter
    def port(self, port: int):
        self.__port = port

    @property
    def graphic_type(self) -> str:
        return self.__graphic_type

    @graphic_type.setter
    def graphic_type(self, graphic_type: str):
        self.__graphic_type = graphic_type

    def __add_open_gl(self):
        data = self.data[self.XML_NAME]["children"]

        if not self.open_gl:
            value = {"gl": {
                "attr": {
                    "enable": "no"
                }
            }}
        else:
            value = {"gl": {
                "attr": {
                    "enable": "yes",
                    "rendernode": self.render_node
                }
            }}
        data.update(value)

    def __add_port(self):
        data = self.data[self.XML_NAME]["attr"]

        if self.port == -1:
            value = {
                "autoport": "yes"
            }
        else:
            value = {
                "autoport": "no",
                "tlsPort": str(self.tls_port),
                "port": str(self.port)
            }
        data.update(value)

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "type": self.graphic_type,
                },
                "children": {
                    "listen": {
                        "attr": {
                            "type": self.listen
                        }
                    },
                    "image": {
                        "attr": {
                            "compression": self.image_compression
                        }
                    }
                }
            }
        })
        self.__add_open_gl()
        self.__add_port()



from pylibvirt.modules.devices import Device


class UsbDevice(Device):
    XML_NAME = "hostdev"

    def __init__(self, vendor_id: str, product_id: str):
        super().__init__(name=self.XML_NAME)
        self.__vendor_id = vendor_id
        self.__product_id = product_id
        self.generate_data()

    @property
    def vendor_id(self) -> str:
        return self.__vendor_id

    @vendor_id.setter
    def vendor_id(self, vendor_id: str):
        self.__vendor_id = vendor_id

    @property
    def product_id(self) -> str:
        return self.__product_id

    @product_id.setter
    def product_id(self, product_id: str):
        self.__product_id = product_id

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "type": "usb",
                    "managed": "yes",
                    "mode": "subsystem"
                },
                "children": {
                    "source": {
                        "children": {
                            "vendor": {
                                "attr": {
                                    "id": self.vendor_id
                                }
                            },
                            "product": {
                                "attr": {
                                    "id": self.product_id
                                }
                            }
                        }
                    },
                }
            }
        })



"""Main module."""

import sys
import importlib
import click
import libvirt
import yaml
from rich import print

from pylibvirt.modules.devices.strorage import DEFAULT_BUS
from pylibvirt.modules.devices.strorage import DiskDevice
from pylibvirt.modules import next_dev, get_dev, Domain
from pylibvirt.modules.network import Network
from pylibvirt.modules.storage import DirStorage
from pylibvirt.modules.volume import Volume


class Manager:
    EXCLUDE_CLASS = ['memory', 'boot_order', 'DiskDevice', 'domain_type', 'Feature',
                     'Os']

    def __init__(self, template: str or dict, force_redefine: bool = False):
        self.__template_path = template
        self.__force_redefine = force_redefine
        self.template = self.load_template()
        self.conn = self.connect()
        print(self.template)
        self.list_domain = []  # List of Domain
        self.list_network = []  # List of Network
        self.list_storage = []  # List of Network
        self.list_volume = []  # List of Network
        self.generate_network()
        self.generate_storage()
        self.generate_volume()
        self.generate_domain()

    def print(self):
        print(self.list_domain)
        print(self.list_network)
        print(self.list_storage)
        print(self.list_volume)

    @property
    def force_redefine(self):
        return self.__force_redefine

    @property
    def template_path(self) -> str:
        return self.__template_path

    @template_path.setter
    def template(self, template: str):
        self.__template_path = template

    @staticmethod
    def load_devices(class_name, param):
        module = importlib.import_module('pylibvirt.modules.devices')
        _class = getattr(module, class_name)
        instance = _class(**param)
        return instance

    @staticmethod
    def target_used(domain_disks: [{}]):
        list_target_use = {}

        for disk in domain_disks:
            for _name, config in disk.items():
                bus = config.get('bus', DEFAULT_BUS)
                target = config.get('target', None)
                if target:
                    list_target = list_target_use.get(bus, [])
                    list_target.append(target)
                    list_target_use.update({bus: list_target})
        return list_target_use

    def generate_dom_target(self, domain_disks: [{}]) -> [{}]:
        list_target_use = self.target_used(domain_disks=domain_disks)
        for disk in domain_disks:
            for _name, config in disk.items():
                bus = config.get('bus', DEFAULT_BUS)
                target = config.get('target', None)
                if not target:
                    target = \
                        next_dev(dev_used=list_target_use.get(bus, []),
                                 dev=get_dev(bus=bus))[
                            0]
                    config.update({'target': target})
        return domain_disks

    def load_template(self):
        if isinstance(self.template_path, str):
            with open(self.template_path) as file:
                return yaml.safe_load(file)
        else:
            return self.template_path

    def connect(self):
        uri = self.template['provider']['uri']
        try:
            conn = libvirt.open(str(uri))
            return conn
        except libvirt.libvirtError:
            exit(1)

    def generate_network(self):
        networks = self.template['network'] if 'network' in self.template else []
        for network in networks:
            for name, config in network.items():
                net = Network(name=name, **config)
                if name in self.conn.listNetworks():
                    obj_net = self.conn.networkLookupByName(name)
                    self.list_network.append(obj_net)
                else:
                    self.list_network.append(
                        self.conn.networkCreateXML(net.generate_xml()))

    def generate_storage(self):
        storages = self.template['storage'] if 'storage' in self.template else []
        for storage in storages:
            for name, config in storage.items():
                pool_type = config.pop('pool_type')
                if pool_type == 'dir':
                    dir = DirStorage(pool_name=name, **config)
                    if name in self.conn.listStoragePools():
                        self.list_storage.append(
                            self.conn.storagePoolLookupByName(name))
                    else:
                        self.list_storage.append(
                            self.conn.storagePoolCreateXML(dir.generate_xml()))

    def generate_volume(self):
        volumes = self.template['volume'] if 'volume' in self.template else []
        for volume in volumes:
            for _name, config in volume.items():
                pool = config.pop('pool')
                clone = config.pop('clone', None)
                vol = Volume(**config)
                pool = self.conn.storagePoolLookupByName(pool)
                if config.get('name') in pool.listVolumes():
                    self.list_volume.append(pool.storageVolLookupByName(
                        config.get('name')))
                else:
                    if not clone:
                        self.list_volume.append(pool.createXML(vol.generate_xml()))
                    else:
                        vol_to_clone = pool.storageVolLookupByName(clone)
                        self.list_volume.append(pool.createXMLFrom(vol.generate_xml(),
                                                                   vol_to_clone, 0))

    def generate_disk(self, domain):
        devices = list()
        disk_devices = domain['DiskDevice']
        disk_devices = self.generate_dom_target(domain_disks=disk_devices)

        for disk in disk_devices:
            for _name, config in disk.items():
                pool = config.pop('pool', 'default')
                vol_name = config.pop('volume')
                pool = self.conn.storagePoolLookupByName(pool)
                volume = pool.storageVolLookupByName(vol_name)
                devices.append(DiskDevice(volume=volume, **config))
        return devices

    def generate_devices(self, domain):
        devices = []
        for device, data in domain.items():
            if device not in self.EXCLUDE_CLASS:
                if isinstance(data, list):
                    for conf in data:
                        for _key, value in conf.items():
                            devices.append(
                                self.load_devices(class_name=device, param=value))
                else:
                    devices.append(
                        self.load_devices(class_name=device, param=data))

        return devices

    def create_dom(self, name, config):
        devices = []
        devices = devices + self.generate_disk(domain=config)
        devices = devices + self.generate_devices(domain=config)
        boot_order = config.get('boot_order', ['hd'])
        domain_type = config.get('domain_type', 'kvm')
        memory = config.get('memory')
        feature = config.get('Feature', None)
        os = config.get('Os', None)
        domain = Domain(name=name, devices=devices, boot_order=boot_order,
                        domain_type=domain_type, feature=feature, os=os)
        domain.memory = domain.set_memory(**memory)
        print(domain.generate_xml())
        list_domain = [dom.name() for dom in self.conn.listAllDomains()]
        if name in list_domain:
            if self.force_redefine:
                dom = self.conn.lookupByName(name)
                dom.undefine()
                self.conn.defineXML(domain.generate_xml())
            else:
                return self.conn.lookupByName(name)
        else:
            return self.conn.defineXML(domain.generate_xml())

    def generate_domain(self):
        domains = self.template['domain'] if 'domain' in self.template else []
        for domain in domains:
            for name, config in domain.items():
                self.list_domain.append(self.create_dom(name=name, config=config))


@click.command()
@click.option('--template', '-t', help='Path to the template file', required=True,
              default="./template/template.yml")
@click.option('--force-redefine', '-f',
              help='Force to redefine XML Element if it already exist', is_flag=True)
def main(args=None, **kwargs):
    """Console script for pylibvirt_."""
    click.echo("Replace this message by putting your code into "
               "pylibvirt_.cli.main")
    click.echo("See click documentation at https://click.palletsprojects.com/")
    Manager(template=kwargs['template'], force_redefine=kwargs['force_redefine'])


if __name__ == "__main__":
    sys.exit(main())



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['pylibvirt', 'pylibvirt.modules', 'pylibvirt.modules.devices']

package_data = \
{'': ['*'], 'pylibvirt': ['template/*']}

install_requires = \
['PyYAML>=5.4.1,<6.0.0',
 'click>=8.0.1,<9.0.0',
 'defusedxml>=0.7.1,<0.8.0',
 'libvirt-python>=7.5.0,<8.0.0',
 'rich>=10.6.0,<11.0.0']

entry_points = \
{'console_scripts': ['pylibvirt = pylibvirt.__main__:main']}

setup_kwargs = {
    'name': 'pylibvirt',
    'version': '0.0.4',
    'description': 'Python package to orchestrate libvirt API from yaml declaration file',
    'long_description': "##############\nPython Libvirt\n##############\n\n*****************\nWhat is pylibvirt\n*****************\n\nPylibvirt is a tools based on python-libvirt api.\nThe goal are to create virtual machines, networks, volumes and storage more easily\nusing yaml description file.\n\n\n***************************\nHow to create template file\n***************************\n\nThe template must contain at least the provider.\n\nAfter you can declare just what you need.\n\nIf you just want to create a Storage on your target you can just add storage element.\n\n**All properties used in the yaml are the same as those used in libvirt**\n\n\nProvider\n========\n\n**Currently the only tested provider is Qemu/Kvm**\n\n\nThe URI format is the same as libvirt api.\n\n.. code-block:: yaml\n\n    provider:\n        uri: qemu+ssh://root@ip/system\n\nnetwork\n========\n\n**Mode** *is the type of network, values can be nat, route, open, hostdev and False for\nisolated network*\n\n.. code-block:: yaml\n\n    network:\n        - network_name:\n              mode: nat\n              domain: network\n              ip4_cidr: ipv4_format_cidr\n              dhcp_start: first_ip_to_use\n              dhcp_stop: last_ip_to_use\n\nstorage\n========\n\nOnly *dir* storage have been tested\n\n.. code-block:: yaml\n\n    storage:\n        - storage_name:\n              pool_type: dir\n              path: storage_path\n\n\nvolume\n======\n\npool is the storage name declared in libvirt or previously in yaml.\n\nname is the end name set to the volume. It's this name that must used in domain section.\n\nThe _key_ has no effect you can set what you want.\n\nIf you want to clone an existing volume or a previouslly create volume you can\nuse the clone argument with the name of the volume. The two volumes must be in the same\npool.\n\n.. code-block:: yaml\n\n    volume:\n      - _key_:\n          disk_format: qcow2\n          capacity: 30\n          size_unit: G\n          pool: storage_name\n          name: volume_name\n\ndomain\n======\n\n_key_ is the name of the VM, choose what you want.\n\nExcept the following key ['memory', 'boot_order', 'DiskDevice', 'domain_type',\n                    'Feature', 'Os']\n\nall other key must be the name of a class in pylibvirt/modules/devices\n\nThe class is called dynamically with the parameters.\nExample if you want to add Rng module add an item call RngDevice (class in\npylibvirt/modules/devices/rng.py)\n\n.. code-block:: yaml\n\n    RngDevice:\n        - first_rng_device:\n            - arg_class: value\n        - second_rng_device\n            - model: virtio\n            - backend_model: random\n            - host_device: /dev/my_custom_random\n\n\n.. code-block:: yaml\n\n    domain:\n      - _key_:\n          boot_order:\n            - cdrom\n            - hd\n          memory:\n            mem_unit: G\n            max_memory: 4\n          Os:\n            arch: x86_64\n            machine: q35\n            os_type: hvm\n          Feature: # Features list: https://libvirt.org/formatdomain.html#hypervisor-features\n            - acpi\n            - kvm:\n                hidden:\n                  state: 'on'\n                poll-control:\n                  state: 'on'\n          CpuDevice:\n            cpu_model: host\n            model_args:\n              fallback: allow\n            vcpu: 2\n            vcpu_args:\n              placement: static\n          GraphicDevice:\n            - spice_server:\n                graphic_type: spice\n          VideoDevice:\n            - screen:\n                model_type: qxl\n                ram: 66500\n          DiskDevice:\n            - disk:\n                volume: debian-10-2.qcow2\n                driver: qemu\n                bus: scsi\n                pool: data\n            - cdrom:\n                volume: debian-10.10.0-amd64-netinst.iso\n                pool: data\n\n          NetworkInterfaceDevice:\n            - default:\n                net_interface: default\n                net_type: network\n                model: e1000e\n\n\n**************\nHow to install\n**************\n\nRequirements\n============\n\nYou need to install the following packages on your system to install python-libvirt in a virtualenv\n\nFedora\n----------\n\nRPM dependencies\n^^^^^^^^^^^^^^^^\n\n.. code-block:: bash\n\n   dnf install python3-devel pkgconfig libvirt-devel\n\n.. code-block:: bash\n\n   pip install pylibvirt\n\nDebian\n----------\nDEB dependencies\n^^^^^^^^^^^^^^^^\n\n.. code-block:: bash\n\n   apt install python3-dev pkg-config libvirt-dev\n\n.. code-block:: bash\n\n   pip install pylibvirt\n\n***********\nHow to use\n***********\n\nCli usage\n=========\n\n.. code-block:: bash\n\n    pylibvirt -t /path/to/template.yml\n\nUse in python code\n==================\n\nTo use pylibvirt in your python code you can do:\n\ncall manager and set file path\n\n.. code-block:: python\n\n    import pylibvirt\n    pylibvirt.Manager(template='path_to_file')\n\n\nor call manager and directly pass template object\n\n.. code-block:: python\n\n    import pylibvirt\n    pylibvirt.Manager(template=[yaml object])\n",
    'author': 'Sevolith',
    'author_email': 'contactsevolith@gmail.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': 'https://gitlab.com/Sevolith/Python-Libvirt/',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.6,<4.0',
}


setup(**setup_kwargs)



from collections import defaultdict

from pylibvirt.modules.devices.xml import XmlGenerator

data_dict = defaultdict(list)


class Device(XmlGenerator):

    def __init__(self, name):
        self.root = ['devices']
        self.xml_name = name
        self.data = defaultdict(list)
        self.data.update({
            name: {

            }
        })

    def __str__(self):
        return self.generate_xml()

    def generate_xml(self) -> str:
        return self.generate_xml_element(data=self.data)

    def generate_data(self):
        raise NotImplementedError

    def update_data(self, data):
        self.data[self.xml_name]["children"].update(data)



from pylibvirt.modules.devices import Device


class ControllerDevice(Device):
    XML_NAME = "controller"

    def __init__(self, controller_type: str = "scsi", model: str = "virtio-scsi"):
        super().__init__(name=self.XML_NAME)
        self.__controller_type = controller_type
        self.__model = model
        self.generate_data()

    @property
    def model(self) -> str:
        return self.__model

    @model.setter
    def model(self, model: str):
        self.__model = model

    @property
    def controller_type(self) -> str:
        return self.__controller_type

    @controller_type.setter
    def controller_type(self, controller_type: str):
        self.__controller_type = controller_type

    def generate_data(self):
        self.data.update({
            self.XML_NAME: {
                "attr": {
                    "model": self.model,
                    "type": self.controller_type,
                }}
        })
